import logging
import os
import signal
import sys
import time
import traceback
from dataclasses import dataclass
from queue import Queue
from typing import Dict, Tuple, Any

import numpy as np

from era_5g_interface.channels import CallbackInfoServer, ChannelType, DATA_NAMESPACE, DATA_ERROR_EVENT
from era_5g_interface.dataclasses.control_command import ControlCommand, ControlCmdType
from era_5g_interface.interface_helpers import HeartBeatSender, MIDDLEWARE_REPORT_INTERVAL, RepeatedTimer
from era_5g_interface.task_handler_internal_q import TaskHandlerInternalQ
from era_5g_server.server import NetworkApplicationServer
from fcw_core.yolo_detector import YOLODetector
from fcw_service.collision_worker import CollisionWorker

logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
logger = logging.getLogger("FCW interface")

# Port of the 5G-ERA Network Application's server.
NETAPP_PORT = int(os.getenv("NETAPP_PORT", 5896))
# Input queue size.
NETAPP_INPUT_QUEUE = int(os.getenv("NETAPP_INPUT_QUEUE", 1))
# Event name for image error.
IMAGE_ERROR_EVENT = str("image_error")


@dataclass
class TaskAndWorker:
    """Class for task and worker."""

    task: TaskHandlerInternalQ
    worker: CollisionWorker


class Server(NetworkApplicationServer):
    """FCW server receives images from clients, manages tasks and workers, sends results to clients."""

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Constructor.

        Args:
            *args: NetworkApplicationServer arguments.
            **kwargs: NetworkApplicationServer arguments.
        """

        super().__init__(
            callbacks_info={
                "image_h264": CallbackInfoServer(ChannelType.H264, self.image_callback),
                "image_hevc": CallbackInfoServer(ChannelType.HEVC, self.image_callback),
                "image_jpeg": CallbackInfoServer(ChannelType.JPEG, self.image_callback),
            },
            *args,
            **kwargs,
        )

        # List of registered tasks.
        self.tasks: Dict[str, TaskAndWorker] = dict()

        self.heart_beat_sender = HeartBeatSender()
        heart_beat_timer = RepeatedTimer(MIDDLEWARE_REPORT_INTERVAL, self.heart_beat)
        heart_beat_timer.start()

    def heart_beat(self):
        """Heart beat generation and sending."""

        latencies = []
        for task_and_worker in self.tasks.values():
            latencies.extend(task_and_worker.worker.latency_measurements.get_latencies())
        avg_latency = 0
        if len(latencies) > 0:
            avg_latency = float(np.mean(np.array(latencies)))

        queue_size = NETAPP_INPUT_QUEUE
        queue_occupancy = 1  # TODO: Compute for every worker?

        self.heart_beat_sender.send_application_heart_beat(
            avg_latency=avg_latency,
            queue_size=queue_size,
            queue_occupancy=queue_occupancy,
            current_robot_count=len(self.tasks),
        )

    def image_callback(self, sid: str, data: Dict[str, Any]) -> None:
        """Allows to receive decoded image using the websocket transport.

        Args:
            sid (str): Namespace sid.
            data (Dict[str, Any]): Data dict including decoded frame (data["frame"]) and send timestamp
                (data["timestamp"]).
        """

        eio_sid = self._sio.manager.eio_sid_from_sid(sid, DATA_NAMESPACE)

        if eio_sid not in self.tasks:
            logger.error(f"Non-registered client {eio_sid} tried to send data")
            self.send_data({"message": "Non-registered client tried to send data"}, DATA_ERROR_EVENT, sid=sid)
            return

        task = self.tasks[eio_sid].task
        task.store_data({"timestamp": data["timestamp"], "recv_timestamp": time.perf_counter_ns()}, data["frame"])

    def command_callback(self, command: ControlCommand, sid: str) -> Tuple[bool, str]:
        """Process initialization control command - create task, worker and start the worker.

        Args:
            command (ControlCommand): Control command to be processed.
            sid (str): Namespace sid.

        Returns:
            (initialized (bool), message (str)): If False, initialization failed.
        """

        eio_sid = self.get_eio_sid_of_control(sid)

        logger.info(f"Control command {command} processing: session id: {sid}")

        if command and command.cmd_type == ControlCmdType.INIT:
            # Check that initialization has not been called before.
            if eio_sid in self.tasks:
                logger.error(f"Client attempted to call initialization multiple times")
                self.send_command_error("Initialization has already been called before", sid)
                return False, "Initialization has already been called before"

            args = command.data
            config = {}
            camera_config = {}
            fps = 30
            viz = True
            viz_zmq_port = 5558
            if args:
                config = args.get("config", config)
                camera_config = args.get("camera_config", camera_config)
                fps = args.get("fps", fps)
                viz = args.get("viz", viz)
                viz_zmq_port = args.get("viz_zmq_port", viz_zmq_port)
                logger.info(f"Config: {config}")
                logger.info(f"Camera config: {camera_config}")
                logger.info(f"ZeroMQ visualization: {viz}, port: {viz_zmq_port}")

            # Queue with received images.
            image_queue = Queue(NETAPP_INPUT_QUEUE)

            task = TaskHandlerInternalQ(image_queue)

            try:
                # Create worker.
                worker = CollisionWorker(
                    image_queue,
                    lambda results: self.send_data(data=results, event="results", sid=self.get_sid_of_data(eio_sid)),
                    config,
                    camera_config,
                    fps,
                    viz,
                    viz_zmq_port,
                    name=f"Collision Worker {eio_sid}",
                    daemon=True,
                )
            except Exception as ex:
                logger.error(f"Failed to create CollisionWorker: {repr(ex)}")
                logger.error(traceback.format_exc())
                self.send_command_error(f"Failed to create CollisionWorker: {repr(ex)}", sid)
                return False, f"Failed to create CollisionWorker: {repr(ex)}"

            self.tasks[eio_sid] = TaskAndWorker(task, worker)
            self.tasks[eio_sid].worker.start()
            t0 = time.perf_counter_ns()
            while True:
                if self.tasks[eio_sid].worker.is_alive():
                    break
                if time.perf_counter_ns() > t0 + 5 * 1.0e9:
                    logger.error(f"Timed out to start worker, eio_sid {eio_sid}, sid {sid}")
                    return False, f"Timed out to start worker"

            logger.info(f"Task handler and worker created and started: {eio_sid}")

        logger.info(
            f"Control command applied, eio_sid {eio_sid}, sid {sid}, "
            f"results sid {self.get_sid_of_data(eio_sid)}, command {command}"
        )
        return True, (
            f"Control command applied, eio_sid {eio_sid}, sid {sid}, results sid"
            f" {self.get_sid_of_data(eio_sid)}, command {command}"
        )

    def disconnect_callback(self, sid: str) -> None:
        """Called with client disconnection - deletes task and worker.

        Args:
            sid (str): Namespace sid.
        """

        eio_sid = self.get_eio_sid_of_data(sid)
        task_and_worker = self.tasks.get(eio_sid)
        if task_and_worker:
            task_and_worker.worker.stop()
            task_and_worker.worker.join()
            del self.tasks[eio_sid]
            del task_and_worker
            logger.info(f"Task handler and worker deleted: {eio_sid}")

        logger.info(f"Client disconnected from {DATA_NAMESPACE} namespace, eio_sid {eio_sid}, sid {sid}")


def main():
    """Main function."""

    # parser = argparse.ArgumentParser(description='Forward Collision Warning Service')
    # args = parser.parse_args()

    logger.info(f"The size of the queue set to: {NETAPP_INPUT_QUEUE}")

    logger.info("Initializing default object detector for faster first startup")
    detector = YOLODetector.from_dict({})
    del detector

    server = Server(port=NETAPP_PORT, host="0.0.0.0")

    try:
        server.run_server()
    except KeyboardInterrupt:
        logger.info("Terminating ...")


if __name__ == "__main__":
    main()



import time
from dataclasses import asdict
from multiprocessing import Queue
from queue import Empty
from threading import Thread, Event
from typing import Callable, Any

import zmq

from era_5g_interface.interface_helpers import LatencyMeasurements
from fcw_core.detection import *
from fcw_core.sort import Sort, KalmanBoxTracker
from fcw_core.yolo_detector import YOLODetector
from fcw_core_utils.collision import *

logger = logging.getLogger(__name__)


class CollisionWorker(Thread):
    """FCW worker. Reads data from passed queue, performs FCW processing and returns results using callback."""

    def __init__(
        self,
        image_queue: Queue,
        send_function: Callable[[Dict[str, Any]], None],
        config: Dict,
        camera_config: Dict,
        fps: float,
        viz: bool = False,
        viz_zmq_port: int = 5558,
        **kw,
    ) -> None:
        """Constructor.

        Args:
            image_queue (Queue): The queue with all to-be-processed images.
            send_function (Callable[[Dict], None]): Callback used to send results.
            config (Dict): FCW config.
            camera_config (Dict): Camera config.
            fps (float): Framerate.
            viz (bool): Enable visualization?
            viz_zmq_port (int): Visualization ZeroMQ port.
            **kw: Thread arguments.
        """

        super().__init__(**kw)

        self._stop_event = Event()
        self.image_queue = image_queue
        self._send_function = send_function
        self._frame_id = 0
        self.latency_measurements: LatencyMeasurements = LatencyMeasurements()
        self._viz = viz

        logger.info("Initializing object detector")
        self._detector = YOLODetector.from_dict(config.get("detector", {}))
        logger.info("Initializing image tracker")
        self._tracker = Sort.from_dict(config.get("tracker", {}))
        logger.info("Initializing forward collision guard")
        self._guard = ForwardCollisionGuard.from_dict(config.get("fcw", {}))
        self._guard.dt = 1 / fps
        logger.info("Initializing camera calibration")
        self._camera = Camera.from_dict(camera_config)
        self._config = dict(config=config, camera_config=camera_config)

        # Visualization stuff.
        if self._viz:
            self._context = zmq.Context()
            self._socket = self._context.socket(zmq.PUB)
            logger.info(f"Publishing visualization on ZeroMQ tcp://*:{viz_zmq_port}")
            self._socket.bind("tcp://*:%s" % viz_zmq_port)

    def stop(self) -> None:
        """Set stop event to stop FCW worker."""

        self._stop_event.set()

    def __del__(self) -> None:
        logger.info("Delete object detector")
        del self._detector

    def run(self) -> None:
        """FWC worker loop. Periodically reads images from python internal queue process them."""

        logger.info(f"{self.name} thread is running.")

        while not self._stop_event.is_set():
            # Get image and metadata from input queue.
            metadata: Dict[str, Any]
            image: np.ndarray
            try:
                metadata, image = self.image_queue.get(block=True, timeout=1)
            except Empty:
                continue
            # Store timestamp before processing.
            metadata["timestamp_before_process"] = time.perf_counter_ns()
            self._frame_id += 1
            # logger.info(f"Worker received frame id: {self.frame_id} {metadata['timestamp']}")
            try:
                detections = self._process_image(image)
                # Store timestamp after processing.
                metadata["timestamp_after_process"] = time.perf_counter_ns()
                # Generate results.
                results = self._generate_results(detections, metadata)
                # Send results via the provided callback.
                self._send_function(results)

                if self._viz:
                    # If visualisation is enabled, send image with results over ZeroMQ.
                    self._send_image_with_results(image, results)

            except Exception as ex:
                logger.error(f"Exception with image processing ({type(ex)}): {repr(ex)}")

        logger.info(f"{self.name} thread is stopping.")

    def _process_image(self, image: np.ndarray) -> Dict[int, KalmanBoxTracker]:
        """Process image by FCW.

        Args:
            image (np.ndarray): Image to be processed.

        Returns:
            Dictionary of KalmanBoxTrackers.
        """

        # Detect object in image.
        detections = self._detector.detect(image)
        # Get bounding boxes as numpy array.
        detections = detections_to_numpy(detections)
        # Update state of image trackers.
        self._tracker.update(detections)
        # Represent trackers as dict, {tid: KalmanBoxTracker, ...}.
        tracked_objects: Dict[int, KalmanBoxTracker] = {
            t.id: t for t in self._tracker.trackers if t.hit_streak > self._tracker.min_hits and t.time_since_update < 1
        }
        # Get 3D locations of objects.
        ref_points = get_reference_points(tracked_objects, self._camera, is_rectified=True)
        # Update state of objects in world.
        self._guard.update(ref_points)

        return tracked_objects

    def _send_image_with_results(self, image: np.ndarray, results: Dict[str, Any]) -> None:
        """Send a numpy array with metadata.

        Args:
            image (np.ndarray): Image.
            results (Dict[str, Any]): FCW results.
        """

        md = dict(dtype=str(image.dtype), shape=image.shape, results=dict(results, config=self._config))
        self._socket.send_json(md, 0 | zmq.SNDMORE)
        self._socket.send(image, 0, copy=True, track=False)

    def _generate_results(
        self, tracked_objects: Dict[int, KalmanBoxTracker], metadata: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Generates the results.

        Args:
            tracked_objects (Dict[int, KalmanBoxTracker]): The results of the detection.
            metadata (Dict[str, Any]): 5G-ERA Network Application specific metadata related to processed image.

        Returns:
            Dictionary of results.
        """

        # Get list of current offenses.
        dangerous_objects = self._guard.dangerous_objects()
        dangerous_detections = dict()

        # Get object statuses.
        object_statuses = list(self._guard.label_objects(include_distant=False))

        if tracked_objects is not None:
            for tid, t in tracked_objects.items():
                x1, y1, x2, y2 = t.get_state()[0]
                det = dict()
                det["bbox"] = [x1, y1, x2, y2]
                det["dangerous_distance"] = 0
                det["age"] = t.age
                det["hit_streak"] = t.hit_streak
                det["class"] = t.label
                det["class_name"] = self._detector.model.names[t.label]

                if tid in dangerous_objects.keys():
                    dist = Point(dangerous_objects[tid].location).distance(self._guard.vehicle_zone)
                    det["dangerous_distance"] = dist
                dangerous_detections[tid] = det

            # Make object statuses serializable - convert from shapely types.
            for i in range(len(object_statuses)):
                object_statuses[i].location = object_statuses[i].location.coords[0]
                object_statuses[i].path = [pts for pts in object_statuses[i].path.coords]
            object_statuses = [asdict(object_status) for object_status in object_statuses]

            return {
                "timestamp": metadata.get("timestamp", 0),
                "recv_timestamp": metadata.get("recv_timestamp", 0),
                "timestamp_before_process": metadata["timestamp_before_process"],
                "timestamp_after_process": metadata["timestamp_after_process"],
                "send_timestamp": time.perf_counter_ns(),
                "dangerous_detections": dangerous_detections,
                "objects": object_statuses,
            }
